"""
Copyright 2016 Workiva, INC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Entry points for goverge."""
import argparse
import os
import shutil
from subprocess import PIPE
from subprocess import Popen
import sys


from coverage import generate_coverage
from reports import compile_reports
from reports import get_coverage_reports


def delete_folder(folder):
    """Delete a folder using the given path, ignoring OSErrors"""
    try:
        shutil.rmtree(folder)
    except OSError:
        return


def get_project_package(project_root, project_import):
    """
    Get the project package from either the import passed in or the project
    root using go list

    :type project_root: string
    :param project_root: The location of the project root
    :type project_import: string
    :param project_import: The import path for the project root
    :rtype: string
    :return: The project package import path
    """
    if not project_import:

        project_import, _ = Popen(
            ["go", "list"],
            stdout=PIPE,
            cwd=project_root
        ).communicate()
    return project_import.replace("'", "")


def main():
    """Main entry point into goverge."""

    os.environ["GORACE"] = "halt_on_error=1"

    args = _parse_args(sys.argv[1:])

    goverge(args)


def goverge(options):
    """
    Goverge command-line interface
    :type options: argparse.Namespace
    :param options: Command-line arguments to control Goverge
    """

    delete_folder("reports")
    os.mkdir("./reports")

    if options.xml:
        try:
            os.mkdir(options.xml_dir)
        except OSError:
            pass

    project_root = os.getcwd()

    project_package = get_project_package(project_root, options.project_import)

    if options.test_path:
        sub_dirs = options.test_path

    else:
        sub_dirs = [x[0] for x in os.walk(project_root)
                    if "/." not in x[0] and "Godeps" not in x[0]]

    generate_coverage(
        sub_dirs, project_package, project_root, options.godep, options.short,
        options.xml, options.xml_dir, options.race, options.tag)

    reports = get_coverage_reports("./reports")

    compile_reports(reports)

    if options.html:
        Popen(
            ["go", "tool", "cover", "--html=test_coverage.txt"],
            stdout=PIPE,
            cwd=project_root
        ).communicate()


def _parse_args(argv):
    """
    Parse command-line arguments.

    :type argv: list
    :param argv: Command-line arguments to parse
    :rtype: argparse.Namespace
    :return: Parsed arguments
    """

    p = argparse.ArgumentParser(prog='goverge')
    p.add_argument(
        '--godep',
        action='store_true',
        default=False,
        help=(
            'Run coverage using the projects godep files.'))

    p.add_argument(
        '--html',
        action='store_true',
        default=False,
        help=(
            "View a html report of the coverage file that is generated."
        )
    )

    p.add_argument(
        '--race',
        action='store_true',
        default=False,
        help=(
            "Run tests using the -race flag"
        )
    )

    p.add_argument(
        '--project_import',
        action='store',
        help=(
            "The import path of the project. leaving this blank will get the "
            "project name Using go list but in some cases that doesn't work "
            "and needs to be manually entered. "
            "example: github.com/Workiva/goverge"
        )
    )

    p.add_argument(
        '--short',
        action='store_true',
        default=False,
        help=(
            'Run coverage using the -short flag'
        )
    )

    p.add_argument(
        '--tag',
        action='store',
        help=(
            "Use an optional build tag when running tests."
        )
    )

    p.add_argument(
        '--test_path',
        default=None,
        action='append',
        help=(
            'Path(s) to a specific package to get the coverage on\n'
            'Example: --test_path path/one --test_path path/two'
        )
    )

    p.add_argument(
        '--xml',
        action='store_true',
        default=False,
        help=(
            "Generate xml reports of test runs, assumes that go2xunit is "
            "installed"
        )
    )

    p.add_argument(
        '--xml_dir',
        action='store',
        default="xml_reports/",
        help=(
            "The location to put the xml reports that are generated."
        )
    )

    return p.parse_args(argv)



"""
Copyright 2016 Workiva, INC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import glob
import os
import subprocess
import sys
import threading


def check_failed(return_code):
    """Check if the test run failed and exit with the correct code if it did.
    :type return_code: int
    :param return_code: The return code of a test run
    """
    if return_code != 0:
        os._exit(return_code)


def generate_coverage(
        packages, project_package, project_root, godep, short, xml, xml_dir,
        race, tag):
    """ Generate the coverage for a list of packages.

    :type package: list
    :param package: Packages to generate coverage for
    :type project_package: string
    :param project_package: The package name of the base of the package
    :type project_root: string
    :param project_root: The project root path
    :type godep: bool
    :param godep: If coverage should run with godep
    :type short: bool
    :param short: If coverage should run with the short flag
    :type xml: bool
    :param xml: If xml test output should be created
    :type xml_dir: string
    :param xml_dir: The location that the xml reports should be made
    :type race: bool
    :param race: If the race flag should be used or not
    :type tag: string
    :param tag: A custom build tag to use when running go test
    """

    max_threads = 4
    threads = []
    while threads or packages:

        if len(threads) < max_threads and packages:
            package = packages.pop()

            # If there isn't a test file in the package we skip it.
            if not glob.glob(u"{0}/*_test.go".format(package)):
                continue

            t = threading.Thread(target=generate_package_coverage, args=(
                package,
                project_package,
                package.replace("/", "_").replace(".", ""),
                project_root,
                godep,
                short,
                xml,
                xml_dir,
                race,
                tag,
            ))
            t.daemon = True
            t.start()
            threads.append(t)

        else:
            threads = [thread for thread in threads if thread.is_alive()]


def generate_package_coverage(
        test_path, project_package, test_package, project_root, godep, short,
        xml, xml_dir, race, tag):
    """ Generates the coverage report for a package.

    :type test_path: string
    :param test_path: Path that is being tested
    :type project_package: string
    :param project_package: The package name of the base of the package
    :type test_package: string
    :param test_package: The name of the package under test
    :type project_root: string
    :param project_root: The project root path
    :type godep: bool
    :param godep: If godep should be used when running tests
    :type xml: bool
    :param xml: If xml test output should be created
    :type xml_dir: string
    :param xml_dir: The location that the xml reports should be made
    :type race: bool
    :param race: If the race flag should be used or not
    :type tag: string
    :param tag: A custom build tag to use when running go test
    """

    # Get the dependencies of the package we are testing
    package_deps = get_package_deps(project_package, test_path)

    options = [
        "go", "test", '-covermode=count',
        u"-coverprofile={0}/reports/{1}.txt".format(
            project_root, test_package),
        u"-coverpkg={0}".format(",".join(package_deps))]

    if godep:
        options = ["godep"] + options

    if short:
        options.append("-short")

    if race:
        options.append("-race")

    if tag:
        options.append("-tag {}".format(tag))

    if xml:
        return generate_xml(xml_dir + test_package, options, test_path)

    # Generate the coverage report for each of the package dependencies
    check_failed(subprocess.call(options, cwd=test_path))


def generate_xml(output_loc, options, test_path):
    """
    Run tests and generate a xml report

    :type output_loc: string
    :param output_loc: The location that the ouput file will be created
    :type options: list
    :param options: The command line options to run the tests with
    :type test_path: string
    :param test_path: Path that is being tested
    """

    output_file = output_loc + ".out"

    with open(output_file, "w") as out_file:
        p = subprocess.Popen(
            options + '-v', cwd=test_path, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        for line in p.stdout:
            sys.stdout.write(line)
            out_file.write(line)
        p.communicate()
    subprocess.call([
        "go2xunit", "-input", output_file, "-output",
        output_loc + ".xml"])
    os.remove(output_file)
    check_failed(p.returncode)


def get_package_deps(project_package, test_path):
    """ Gets the packages dependencies that are part of the project.

    :type project_package: string
    :param project_package: The base package of the project
    :type test_path: string
    :param test_path: The path of the package that is under test
    :rtype: list
    :return: Project dependencies
    """

    output, _ = subprocess.Popen(
        ["go", "list", "-f", "'{{.Deps}}'"],
        stdout=subprocess.PIPE, cwd=test_path).communicate()

    test_output, _ = subprocess.Popen(
        ["go", "list", "-f", "'{{.TestImports}}'"],
        stdout=subprocess.PIPE, cwd=test_path).communicate()

    output = test_output.split() + output.split()

    package_deps = [
        package.replace("]", "").replace("[", "").replace("'", "")
        for package in list(set(output))
        if project_package.split("\n")[0] in package]

    package_deps.append(".")

    return package_deps
